#ifndef LiquidCrystal3WI_h
#define LiquidCrystal3WI_h

#include <inttypes.h>
#include "Print.h"

// commands
#define LCD_CLEARDISPLAY 0x01
#define LCD_RETURNHOME 0x02
#define LCD_ENTRYMODESET 0x04
#define LCD_DISPLAYCONTROL 0x08
#define LCD_CURSORSHIFT 0x10
#define LCD_FUNCTIONSET 0x20
#define LCD_SETCGRAMADDR 0x40
#define LCD_SETDDRAMADDR 0x80

// flags for display entry mode
#define LCD_ENTRYRIGHT 0x00
#define LCD_ENTRYLEFT 0x02
#define LCD_ENTRYSHIFTINCREMENT 0x01
#define LCD_ENTRYSHIFTDECREMENT 0x00

// flags for display on/off control
#define LCD_DISPLAYON 0x04
#define LCD_DISPLAYOFF 0x00
#define LCD_CURSORON 0x02
#define LCD_CURSOROFF 0x00
#define LCD_BLINKON 0x01
#define LCD_BLINKOFF 0x00

// flags for display/cursor shift
#define LCD_DISPLAYMOVE 0x08
#define LCD_CURSORMOVE 0x00
#define LCD_MOVERIGHT 0x04
#define LCD_MOVELEFT 0x00

// flags for function set
#define LCD_4BITMODE 0x00
#define LCD_2LINE 0x08
#define LCD_1LINE 0x00
#define LCD_5x10DOTS 0x04
#define LCD_5x8DOTS 0x00

#define SEND_BIT(DATA_PIN, CLOCK_PIN, BIT) {digitalWrite(DATA_PIN, BIT); \
											digitalWrite(CLOCK_PIN, HIGH); \
											digitalWrite(CLOCK_PIN, LOW); \
											digitalWrite(DATA_PIN, LOW);}

class LiquidCrystal3WI : public Print {
public:
  LiquidCrystal3WI(uint8_t dataPin, uint8_t clockPin, uint8_t latchPin);
  LiquidCrystal3WI(uint8_t dataPin, uint8_t clockPin, uint8_t latchPin, bool usePage2);
  virtual ~LiquidCrystal3WI() {};

  void init(uint8_t dataPin, uint8_t clockPin, uint8_t latchPin, bool usePage2);
    
  void begin(uint8_t cols, uint8_t rows, uint8_t charsize = LCD_5x8DOTS);

  void clear();
  void home();

  void noDisplay();
  void display();
  void noBlink();
  void blink();
  void noCursor();
  void cursor();
  void scrollDisplayLeft();
  void scrollDisplayRight();
  void leftToRight();
  void rightToLeft();
  void autoscroll();
  void noAutoscroll();

  void setBacklight(bool on);

  void setRowOffsets(int row1, int row2, int row3, int row4);
  void createChar(uint8_t, uint8_t[]);
  void setCursor(uint8_t, uint8_t); 
  virtual size_t write(uint8_t);
  void command(uint8_t);
  
  void printUTF8(const char *utf8String);
  void printUTF8(uint16_t ch);

  using Print::write;
private:
  void send(uint8_t, uint8_t);
  void write4bits(uint8_t);
  void pulseEnable();

  void setupDisplayPins(uint8_t enablePin, uint8_t resetPin, uint8_t rwPin, uint8_t data, bool setData);

  void setupRegisterData(uint8_t enablePin, uint8_t resetPin, uint8_t rwPin, uint8_t data, bool setData);
  void setupDisplayPins();

  uint8_t _dataPin;
  uint8_t _clockPin;
  uint8_t _latchPin;

  uint8_t _registerData;

  uint8_t _displayfunction;
  uint8_t _displaycontrol;
  uint8_t _displaymode;

  uint8_t _initialized;

  uint8_t _numlines;
  uint8_t _row_offsets[4];

  bool _backlignt;
  bool _usePage2;
};

#endif
